
enum {CYCL, EDGE} ;

class RemapState {
  const int typeLen ;
  const int rank ;
  Range**  locRngs ;
  Stride** locStrs ;
  Range**  remRngs ;
  Stride** remStrs ;
  ProcessGroup* remProcGrp ;
  DataExchange* exchange ;
  Array1<int>** subscripts ;
  int* exts ;
  int* locSteps ;
  int* remSteps ;

  void skewGetSubrange(char* locData, int locStep,
                       int remOffset, Range* remRng, Stride* remStr) ;
  void remapGetSubrange(char* locData, int locStep,
                        int remOffset, Range* remRng, Stride* remStr,
                        const int d) ;
  void shiftGetSubrange(char* locData, int locStep,
			int remOffset, Range* remRng, Stride* remStr,
			const int d,
			const int dim, const int shift, const int mode) ;
public :
  RemapState(const int _typeLen, const int _rank,
	     Range** _locRngs, Stride** _locStrs,
	     Range** _remRngs, Stride** _remStrs,
	     ProcessGroup* _remProcGrp,
	     DataExchange* _exchange,
             Array1<int>** _subscripts) ;
  ~RemapState() ;

  void remapLoop(char* locData, int remOffset, int d) ;
  void vecGatherLoop(char* locData, int remOffset, int r, int s) ;
  void vecScatterLoop(char* locData, int remOffset, int r, int s) ;
  void shiftLoop(char* locData, int remOffset, int d,
                 const int dim, const int shift, const int mode) ;
  void skewLoop(char* locData, int remOffset, const int d, const int r,
                const int dim, char* shfData, Stride** shfStrs,
                const int mode) ;
} ;


class CopyState {
  const int typeLen ;
  const int rank ;
  Range**  rngs ;
  Stride** dstStrs ;
  Stride** srcStrs ;
public :
  CopyState(const int _typeLen, const int _rank,
	    Range** _rngs,
            Stride** _dstStrs, Stride** _srcStrs) ;

  void copyLoop(char* dstData, char* srcData, const int d) ;
} ;


class InitState {
  const int typeLen ;
  const int rank ;
  Range**  rngs ;
  Stride** strs ;
public :
  InitState(const int _typeLen, const int _rank,
	    Range** _rngs, Stride** _strs) ;

  void initLoop(char* dstData, char* srcData, const int d) ;
} ;


void copy(Frame* dst, Frame* src) ;
void remap(Frame* dst, Frame* src) ;
void vecGather(Frame* dst, Frame* src, Array1<int>** subscripts) ;
void vecScatter(Frame* src, Frame* dst, Array1<int>** subscripts) ;
void shift(Frame* result, Frame* arr,
           const int shift, const int dim, const int mode) ;
void skew(Frame* result, Frame* arr, Frame* shift,
          const int dim, const int mode) ;
void gather(Frame* dst, Frame* src, Array<int>** subscripts) ;
//void reshape(Frame* dst, Frame* src) ;
void reshape(Frame* dst, Frame* src, Frame* pad) ;


inline void remap(Frame& dst, Frame& src) {remap(&dst, &src) ; }

